@@ -21,9 +21,9 @@ OTHER DEALINGS IN THE SOFTWARE.
 """
 from __future__ import absolute_import
 
-import json
 import sys
 from collections import OrderedDict, namedtuple
+from distutils.util import strtobool
 from functools import partial
 from itertools import chain
 from types import ModuleType
@@ -37,7 +37,6 @@ from hug import introspect
 from hug._async import asyncio, ensure_future
 from hug._version import current
 
-
 INTRO = """
 /#######################################################################\\
           `.----``..-------..``.----.
@@ -99,6 +98,23 @@ class HTTPInterfaceAPI(InterfaceAPI):
         """Returns the active not found handler"""
         return getattr(self, '_not_found', self.base_404)
 
+    def urls(self):
+        """Returns a generator of all URLs attached to this API"""
+        for base_url, mapping in self.routes.items():
+            for url, _ in mapping.items():
+                yield base_url + url
+
+    def handlers(self):
+        """Returns all registered handlers attached to this API"""
+        used = []
+        for base_url, mapping in self.routes.items():
+            for url, methods in mapping.items():
+                for method, versions in methods.items():
+                    for version, handler in versions.items():
+                        if not handler in used:
+                            used.append(handler)
+                            yield handler
+
     def input_format(self, content_type):
         """Returns the set input_format handler for the given content_type"""
         return getattr(self, '_input_format', {}).get(content_type, hug.defaults.input_format.get(content_type, None))
@@ -160,11 +176,12 @@ class HTTPInterfaceAPI(InterfaceAPI):
         for middleware in (http_api.middleware or ()):
             self.add_middleware(middleware)
 
-        for version, handler in getattr(self, '_exception_handlers', {}).items():
-            for exception_type, exception_handler in handler.items():
-                target_exception_handlers = http_api.exception_handlers(version) or {}
-                if exception_type not in target_exception_handlers:
-                    http_api.add_exception_handler(exception_type, exception_handler, version)
+        for version, handler in getattr(http_api, '_exception_handlers', {}).items():
+            for exception_type, exception_handlers in handler.items():
+                target_exception_handlers = self.exception_handlers(version) or {}
+                for exception_handler in exception_handlers:
+                    if exception_type not in target_exception_handlers:
+                        self.add_exception_handler(exception_type, exception_handler, version)
 
         for input_format, input_format_handler in getattr(http_api, '_input_format', {}).items():
             if not input_format in getattr(self, '_input_format', {}):
@@ -225,11 +242,10 @@ class HTTPInterfaceAPI(InterfaceAPI):
                             doc = version_dict.setdefault(url, OrderedDict())
                             doc[method] = handler.documentation(doc.get(method, None), version=version, prefix=prefix,
                                                                 base_url=router_base_url, url=url)
-
         documentation['handlers'] = version_dict
         return documentation
 
-    def serve(self, port=8000, no_documentation=False, display_intro=True):
+    def serve(self, host='', port=8000, no_documentation=False, display_intro=True):
         """Runs the basic hug development server against this API"""
         if no_documentation:
             api = self.server(None)
@@ -239,8 +255,8 @@ class HTTPInterfaceAPI(InterfaceAPI):
         if display_intro:
             print(INTRO)
 
-        httpd = make_server('', port, api)
-        print("Serving on port {0}...".format(port))
+        httpd = make_server(host, port, api)
+        print("Serving on {0}:{1}...".format(host, port))
         httpd.serve_forever()
 
     @staticmethod
@@ -288,9 +304,9 @@ class HTTPInterfaceAPI(InterfaceAPI):
                                 "Here's a definition of the API to help you get going :)")
             to_return['documentation'] = self.documentation(base_url, self.determine_version(request, False),
                                                             prefix=url_prefix)
-            response.data = json.dumps(to_return, indent=4, separators=(',', ': ')).encode('utf8')
+            response.data = hug.output_format.json(to_return, indent=4, separators=(',', ': '))
             response.status = falcon.HTTP_NOT_FOUND
-            response.content_type = 'application/json'
+            response.content_type = 'application/json; charset=utf-8'
         handle_404.interface = True
         return handle_404
 
@@ -355,11 +371,12 @@ HTTPInterfaceAPI.base_404.interface = True
 
 class CLIInterfaceAPI(InterfaceAPI):
     """Defines the CLI interface specific API"""
-    __slots__ = ('commands', )
+    __slots__ = ('commands', 'error_exit_codes',)
 
-    def __init__(self, api, version=''):
+    def __init__(self, api, version='', error_exit_codes=False):
         super().__init__(api)
         self.commands = {}
+        self.error_exit_codes = error_exit_codes
 
     def __call__(self, args=None):
         """Routes to the correct command line tool"""
@@ -370,7 +387,14 @@ class CLIInterfaceAPI(InterfaceAPI):
             return sys.exit(1)
 
         command = args.pop(1)
-        self.commands.get(command)()
+        result = self.commands.get(command)()
+
+        if self.error_exit_codes and bool(strtobool(result.decode('utf-8'))) is False:
+            sys.exit(1)
+
+    def handlers(self):
+        """Returns all registered handlers attached to this API"""
+        return self.commands.values()
 
     def __str__(self):
         return "{0}\n\nAvailable Commands:{1}\n".format(self.api.doc or self.api.name,
@@ -405,9 +429,10 @@ class ModuleSingleton(type):
 
 class API(object, metaclass=ModuleSingleton):
     """Stores the information necessary to expose API calls within this module externally"""
-    __slots__ = ('module', '_directives', '_http', '_cli', '_context', '_startup_handlers', 'started', 'name', 'doc')
+    __slots__ = ('module', '_directives', '_http', '_cli', '_context', '_context_factory', '_delete_context',
+                 '_startup_handlers', 'started', 'name', 'doc', 'cli_error_exit_codes')
 
-    def __init__(self, module=None, name='', doc=''):
+    def __init__(self, module=None, name='', doc='', cli_error_exit_codes=False):
         self.module = module
         if module:
             self.name = name or module.__name__ or ''
@@ -416,6 +441,7 @@ class API(object, metaclass=ModuleSingleton):
             self.name = name
             self.doc = doc
         self.started = False
+        self.cli_error_exit_codes = cli_error_exit_codes
 
     def directives(self):
         """Returns all directives applicable to this Hug API"""
@@ -430,6 +456,13 @@ class API(object, metaclass=ModuleSingleton):
         self._directives = getattr(self, '_directives', {})
         self._directives[directive.__name__] = directive
 
+    def handlers(self):
+        """Returns all registered handlers attached to this API"""
+        if getattr(self, '_http'):
+            yield from self.http.handlers()
+        if getattr(self, '_cli'):
+            yield from self.cli.handlers()
+
     @property
     def http(self):
         if not hasattr(self, '_http'):
@@ -439,9 +472,25 @@ class API(object, metaclass=ModuleSingleton):
     @property
     def cli(self):
         if not hasattr(self, '_cli'):
-            self._cli = CLIInterfaceAPI(self)
+            self._cli = CLIInterfaceAPI(self, error_exit_codes=self.cli_error_exit_codes)
         return self._cli
 
+    @property
+    def context_factory(self):
+        return getattr(self, '_context_factory', hug.defaults.context_factory)
+
+    @context_factory.setter
+    def context_factory(self, context_factory_):
+        self._context_factory = context_factory_
+
+    @property
+    def delete_context(self):
+        return getattr(self, '_delete_context', hug.defaults.delete_context)
+
+    @delete_context.setter
+    def delete_context(self, delete_context_):
+        self._delete_context = delete_context_
+
     @property
     def context(self):
         if not hasattr(self, '_context'):
