@@ -28,16 +28,17 @@ from collections import OrderedDict
 from functools import lru_cache, partial, wraps
 
 import falcon
+from falcon import HTTP_BAD_REQUEST
+
 import hug._empty as empty
 import hug.api
 import hug.output_format
 import hug.types as types
-from falcon import HTTP_BAD_REQUEST
 from hug import introspect
 from hug._async import asyncio_call
 from hug.exceptions import InvalidTypeData
 from hug.format import parse_content_type
-from hug.types import MarshmallowSchema, Multiple, OneOf, SmartBoolean, Text, text
+from hug.types import MarshmallowInputSchema, MarshmallowReturnSchema, Multiple, OneOf, SmartBoolean, Text, text
 
 
 class Interfaces(object):
@@ -87,7 +88,7 @@ class Interfaces(object):
             if hasattr(transformer, 'from_string'):
                 transformer = transformer.from_string
             elif hasattr(transformer, 'load'):
-                transformer = MarshmallowSchema(transformer)
+                transformer = MarshmallowInputSchema(transformer)
             elif hasattr(transformer, 'deserialize'):
                 transformer = transformer.deserialize
 
@@ -109,7 +110,7 @@ class Interface(object):
     """
     __slots__ = ('interface', '_api', 'defaults', 'parameters', 'required', '_outputs', 'on_invalid', 'requires',
                  'validate_function', 'transform', 'examples', 'output_doc', 'wrapped', 'directives', 'all_parameters',
-                 'raise_on_invalid', 'invalid_outputs')
+                 'raise_on_invalid', 'invalid_outputs', 'map_params', 'input_transformations')
 
     def __init__(self, route, function):
         if route.get('api', None):
@@ -137,6 +138,26 @@ class Interface(object):
             self.all_parameters = set(route['parameters'])
             self.required = tuple([parameter for parameter in self.parameters if parameter not in self.defaults])
 
+        if 'map_params' in route:
+            self.map_params = route['map_params']
+            for interface_name, internal_name in self.map_params.items():
+                if internal_name in self.defaults:
+                    self.defaults[interface_name] = self.defaults.pop(internal_name)
+                if internal_name in self.parameters:
+                    self.parameters = [interface_name if param == internal_name else param for param in self.parameters]
+                if internal_name in self.all_parameters:
+                    self.all_parameters.remove(internal_name)
+                    self.all_parameters.add(interface_name)
+                if internal_name in self.required:
+                    self.required = tuple([interface_name if param == internal_name else param for
+                                           param in self.required])
+
+            reverse_mapping = {internal: interface for interface, internal in self.map_params.items()}
+            self.input_transformations = {reverse_mapping.get(name, name): transform for
+                                          name, transform in self.interface.input_transformations.items()}
+        else:
+            self.input_transformations = self.interface.input_transformations
+
         if 'output' in route:
             self.outputs = route['output']
 
@@ -145,7 +166,7 @@ class Interface(object):
             self.transform = self.interface.transform
 
         if hasattr(self.transform, 'dump'):
-            self.transform = self.transform.dump
+            self.transform = MarshmallowReturnSchema(self.transform)
             self.output_doc = self.transform.__doc__
         elif self.transform or self.interface.transform:
             output_doc = (self.transform or self.interface.transform)
@@ -174,17 +195,26 @@ class Interface(object):
     def outputs(self, outputs):
         self._outputs = outputs  # pragma: no cover - generally re-implemented by sub classes
 
-    def validate(self, input_parameters):
+    def validate(self, input_parameters, context):
         """Runs all set type transformers / validators against the provided input parameters and returns any errors"""
         errors = {}
-        for key, type_handler in self.interface.input_transformations.items():
+
+        for key, type_handler in self.input_transformations.items():
             if self.raise_on_invalid:
                 if key in input_parameters:
-                    input_parameters[key] = type_handler(input_parameters[key])
+                    input_parameters[key] = self.initialize_handler(
+                        type_handler,
+                        input_parameters[key],
+                        context=context
+                    )
             else:
                 try:
                     if key in input_parameters:
-                        input_parameters[key] = type_handler(input_parameters[key])
+                        input_parameters[key] = self.initialize_handler(
+                            type_handler,
+                            input_parameters[key],
+                            context=context
+                        )
                 except InvalidTypeData as error:
                     errors[key] = error.reasons or str(error.message)
                 except Exception as error:
@@ -192,26 +222,24 @@ class Interface(object):
                         errors[key] = error.args[0]
                     else:
                         errors[key] = str(error)
-
-        for require in self.interface.required:
+        for require in self.required:
             if not require in input_parameters:
                 errors[require] = "Required parameter '{}' not supplied".format(require)
         if not errors and getattr(self, 'validate_function', False):
             errors = self.validate_function(input_parameters)
         return errors
 
-    def check_requirements(self, request=None, response=None):
+    def check_requirements(self, request=None, response=None, context=None):
         """Checks to see if all requirements set pass
 
            if all requirements pass nothing will be returned
            otherwise, the error reported will be returned
         """
         for requirement in self.requires:
-            conclusion = requirement(response=response, request=request, module=self.api.module)
+            conclusion = requirement(response=response, request=request, context=context, module=self.api.module)
             if conclusion and conclusion is not True:
                 return conclusion
 
-
     def documentation(self, add_to=None):
         """Produces general documentation for the interface"""
         doc = OrderedDict if add_to is None else add_to
@@ -244,6 +272,24 @@ class Interface(object):
 
         return doc
 
+    def _rewrite_params(self, params):
+        for interface_name, internal_name in self.map_params.items():
+            if interface_name in params:
+                params[internal_name] = params.pop(interface_name)
+
+    @staticmethod
+    def cleanup_parameters(parameters, exception=None):
+        for parameter, directive in parameters.items():
+            if hasattr(directive, 'cleanup'):
+                directive.cleanup(exception=exception)
+
+    @staticmethod
+    def initialize_handler(handler, value, context):
+        try:  # It's easier to ask for forgiveness than for permission
+            return handler(value, context=context)
+        except TypeError:
+            return handler(value)
+
 
 class Local(Interface):
     """Defines the Interface responsible for exposing functions locally"""
@@ -272,10 +318,13 @@ class Local(Interface):
         return self.interface.spec.__module__
 
     def __call__(self, *args, **kwargs):
+        context = self.api.context_factory(api=self.api, api_version=self.version, interface=self)
         """Defines how calling the function locally should be handled"""
+
         for requirement in self.requires:
-            lacks_requirement = self.check_requirements()
+            lacks_requirement = self.check_requirements(context=context)
             if lacks_requirement:
+                self.api.delete_context(context, lacks_requirement=lacks_requirement)
                 return self.outputs(lacks_requirement) if self.outputs else lacks_requirement
 
         for index, argument in enumerate(args):
@@ -287,20 +336,32 @@ class Local(Interface):
                     continue
                 arguments = (self.defaults[parameter], ) if parameter in self.defaults else ()
                 kwargs[parameter] = directive(*arguments, api=self.api, api_version=self.version,
-                                              interface=self)
+                                              interface=self, context=context)
 
         if not getattr(self, 'skip_validation', False):
-            errors = self.validate(kwargs)
+            errors = self.validate(kwargs, context)
             if errors:
                 errors = {'errors': errors}
                 if getattr(self, 'on_invalid', False):
                     errors = self.on_invalid(errors)
                 outputs = getattr(self, 'invalid_outputs', self.outputs)
+                self.api.delete_context(context, errors=errors)
                 return outputs(errors) if outputs else errors
 
-        result = self.interface(**kwargs)
-        if self.transform:
-            result = self.transform(result)
+        if getattr(self, 'map_params', None):
+            self._rewrite_params(kwargs)
+        try:
+            result = self.interface(**kwargs)
+            if self.transform:
+                if hasattr(self.transform, 'context'):
+                    self.transform.context = context
+                result = self.transform(result)
+        except Exception as exception:
+            self.cleanup_parameters(kwargs, exception=exception)
+            self.api.delete_context(context, exception=exception)
+            raise exception
+        self.cleanup_parameters(kwargs)
+        self.api.delete_context(context)
         return self.outputs(result) if self.outputs else result
 
 
@@ -318,14 +379,25 @@ class CLI(Interface):
 
         used_options = {'h', 'help'}
         nargs_set = self.interface.takes_args or self.interface.takes_kwargs
-        self.parser = argparse.ArgumentParser(description=route.get('doc', self.interface.spec.__doc__))
+
+        class CustomArgumentParser(argparse.ArgumentParser):
+            exit_callback = None
+
+            def exit(self, status=0, message=None):
+                if self.exit_callback:
+                    self.exit_callback(message)
+                super().exit(status, message)
+
+        self.parser = CustomArgumentParser(description=route.get('doc', self.interface.spec.__doc__))
         if 'version' in route:
             self.parser.add_argument('-v', '--version', action='version',
                                 version="{0} {1}".format(route.get('name', self.interface.spec.__name__),
                                                          route['version']))
             used_options.update(('v', 'version'))
 
+        self.context_tranforms = []
         for option in use_parameters:
+
             if option in self.directives:
                 continue
 
@@ -388,9 +460,11 @@ class CLI(Interface):
     def outputs(self, outputs):
         self._outputs = outputs
 
-    def output(self, data):
+    def output(self, data, context):
         """Outputs the provided data using the transformations and output format specified for this CLI endpoint"""
         if self.transform:
+            if hasattr(self.transform, 'context'):
+                self.transform.context = context
             data = self.transform(data)
         if hasattr(data, 'read'):
             data = data.read().decode('utf8')
@@ -404,11 +478,18 @@ class CLI(Interface):
 
     def __call__(self):
         """Calls the wrapped function through the lens of a CLI ran command"""
+        context = self.api.context_factory(api=self.api, argparse=self.parser, interface=self)
+
+        def exit_callback(message):
+            self.api.delete_context(context, errors=message)
+        self.parser.exit_callback = exit_callback
+
         self.api._ensure_started()
         for requirement in self.requires:
-            conclusion = requirement(request=sys.argv, module=self.api.module)
+            conclusion = requirement(request=sys.argv, module=self.api.module, context=context)
             if conclusion and conclusion is not True:
-                return self.output(conclusion)
+                self.api.delete_context(context, lacks_requirement=conclusion)
+                return self.output(conclusion, context)
 
         if self.interface.is_method:
             self.parser.prog = "%s %s" % (self.api.module.__name__, self.interface.name)
@@ -417,17 +498,24 @@ class CLI(Interface):
         pass_to_function = vars(known)
         for option, directive in self.directives.items():
             arguments = (self.defaults[option], ) if option in self.defaults else ()
-            pass_to_function[option] = directive(*arguments, api=self.api, argparse=self.parser,
+            pass_to_function[option] = directive(*arguments, api=self.api, argparse=self.parser, context=context,
                                                  interface=self)
+
         for field, type_handler in self.reaffirm_types.items():
             if field in pass_to_function:
-                pass_to_function[field] = type_handler(pass_to_function[field])
+                pass_to_function[field] = self.initialize_handler(
+                    type_handler,
+                    pass_to_function[field],
+                    context=context
+                )
 
         if getattr(self, 'validate_function', False):
             errors = self.validate_function(pass_to_function)
             if errors:
-                return self.output(errors)
+                self.api.delete_context(context, errors=errors)
+                return self.output(errors, context)
 
+        args = None
         if self.additional_options:
             args = []
             for parameter in self.interface.parameters:
@@ -449,11 +537,21 @@ class CLI(Interface):
                     elif add_options_to:
                         pass_to_function[add_options_to].append(option)
 
-            result = self.interface(*args, **pass_to_function)
-        else:
-            result = self.interface(**pass_to_function)
+        if getattr(self, 'map_params', None):
+            self._rewrite_params(pass_to_function)
 
-        return self.output(result)
+        try:
+            if args:
+                result = self.output(self.interface(*args, **pass_to_function), context)
+            else:
+                result = self.output(self.interface(**pass_to_function), context)
+        except Exception as exception:
+            self.cleanup_parameters(pass_to_function, exception=exception)
+            self.api.delete_context(context, exception=exception)
+            raise exception
+        self.cleanup_parameters(pass_to_function)
+        self.api.delete_context(context)
+        return result
 
 
 class HTTP(Interface):
@@ -477,8 +575,7 @@ class HTTP(Interface):
         elif self.transform:
             self._params_for_on_invalid = self._params_for_transform
 
-        if route['versions']:
-            self.api.http.versions.update(route['versions'])
+        self.api.http.versions.update(route.get('versions', (None, )))
 
         self.interface.http = self
 
@@ -501,9 +598,10 @@ class HTTP(Interface):
             self._params_for_transform_state = introspect.takes_arguments(self.transform, *self.AUTO_INCLUDE)
         return self._params_for_transform_state
 
-    def gather_parameters(self, request, response, api_version=None, **input_parameters):
+    def gather_parameters(self, request, response, context, api_version=None, **input_parameters):
         """Gathers and returns all parameters that will be used for this endpoint"""
         input_parameters.update(request.params)
+
         if self.parse_body and request.content_length:
             body = request.stream
             content_type, content_params = parse_content_type(request.content_type)
@@ -526,8 +624,8 @@ class HTTP(Interface):
         for parameter, directive in self.directives.items():
             arguments = (self.defaults[parameter], ) if parameter in self.defaults else ()
             input_parameters[parameter] = directive(*arguments, response=response, request=request,
-                                                    api=self.api, api_version=api_version, interface=self)
-
+                                                    api=self.api, api_version=api_version, context=context,
+                                                    interface=self)
         return input_parameters
 
     @property
@@ -538,13 +636,16 @@ class HTTP(Interface):
     def outputs(self, outputs):
         self._outputs = outputs
 
-    def transform_data(self, data, request=None, response=None):
+    def transform_data(self, data, request=None, response=None, context=None):
+        transform = self.transform
+        if hasattr(transform, 'context'):
+            self.transform.context = context
         """Runs the transforms specified on this endpoint with the provided data, returning the data modified"""
-        if self.transform and not (isinstance(self.transform, type) and isinstance(data, self.transform)):
+        if transform and not (isinstance(transform, type) and isinstance(data, transform)):
             if self._params_for_transform:
-                return self.transform(data, **self._arguments(self._params_for_transform, request, response))
+                return transform(data, **self._arguments(self._params_for_transform, request, response))
             else:
-                return self.transform(data)
+                return transform(data)
         return data
 
     def content_type(self, request=None, response=None):
@@ -596,10 +697,12 @@ class HTTP(Interface):
     def call_function(self, parameters):
         if not self.interface.takes_kwargs:
             parameters = {key: value for key, value in parameters.items() if key in self.all_parameters}
+        if getattr(self, 'map_params', None):
+            self._rewrite_params(parameters)
 
         return self.interface(**parameters)
 
-    def render_content(self, content, request, response, **kwargs):
+    def render_content(self, content, context, request, response, **kwargs):
         if hasattr(content, 'interface') and (content.interface is True or hasattr(content.interface, 'http')):
             if content.interface is True:
                 content(request, response, api_version=None, **kwargs)
@@ -607,7 +710,7 @@ class HTTP(Interface):
                 content.interface.http(request, response, api_version=None, **kwargs)
             return
 
-        content = self.transform_data(content, request, response)
+        content = self.transform_data(content, request, response, context)
         content = self.outputs(content, **self._arguments(self._params_for_outputs, request, response))
         if hasattr(content, 'read'):
             size = None
@@ -632,6 +735,8 @@ class HTTP(Interface):
             response.data = content
 
     def __call__(self, request, response, api_version=None, **kwargs):
+        context = self.api.context_factory(response=response, request=request, api=self.api, api_version=api_version,
+                                           interface=self)
         """Call the wrapped function over HTTP pulling information as needed"""
         if isinstance(api_version, str) and api_version.isdigit():
             api_version = int(api_version)
@@ -642,24 +747,30 @@ class HTTP(Interface):
         else:
             exception_types = self.api.http.exception_handlers(api_version)
             exception_types = tuple(exception_types.keys()) if exception_types else ()
+        input_parameters = {}
         try:
             self.set_response_defaults(response, request)
-
-            lacks_requirement = self.check_requirements(request, response)
+            lacks_requirement = self.check_requirements(request, response, context)
             if lacks_requirement:
                 response.data = self.outputs(lacks_requirement,
                                              **self._arguments(self._params_for_outputs, request, response))
+                self.api.delete_context(context, lacks_requirement=lacks_requirement)
                 return
 
-            input_parameters = self.gather_parameters(request, response, api_version, **kwargs)
-            errors = self.validate(input_parameters)
+            input_parameters = self.gather_parameters(request, response, context, api_version, **kwargs)
+            errors = self.validate(input_parameters, context)
             if errors:
+                self.api.delete_context(context, errors=errors)
                 return self.render_errors(errors, request, response)
 
-            self.render_content(self.call_function(input_parameters), request, response, **kwargs)
-        except falcon.HTTPNotFound:
+            self.render_content(self.call_function(input_parameters), context, request, response, **kwargs)
+        except falcon.HTTPNotFound as exception:
+            self.cleanup_parameters(input_parameters, exception=exception)
+            self.api.delete_context(context, exception=exception)
             return self.api.http.not_found(request, response, **kwargs)
         except exception_types as exception:
+            self.cleanup_parameters(input_parameters, exception=exception)
+            self.api.delete_context(context, exception=exception)
             handler = None
             exception_type = type(exception)
             if exception_type in exception_types:
@@ -676,6 +787,12 @@ class HTTP(Interface):
                 raise exception
 
             handler(request=request, response=response, exception=exception, **kwargs)
+        except Exception as exception:
+            self.cleanup_parameters(input_parameters, exception=exception)
+            self.api.delete_context(context, exception=exception)
+            raise exception
+        self.cleanup_parameters(input_parameters)
+        self.api.delete_context(context)
 
     def documentation(self, add_to=None, version=None, prefix="", base_url="", url=""):
         """Returns the documentation specific to an HTTP interface"""
