@@ -29,10 +29,11 @@ from functools import wraps
 from urllib.parse import urljoin
 
 import falcon
+from falcon import HTTP_METHODS
+
 import hug.api
 import hug.interface
 import hug.output_format
-from falcon import HTTP_METHODS
 from hug import introspect
 from hug.exceptions import InvalidTypeData
 
@@ -41,7 +42,7 @@ class Router(object):
     """The base chainable router object"""
     __slots__ = ('route', )
 
-    def __init__(self, transform=None, output=None, validate=None, api=None, requires=(), **kwargs):
+    def __init__(self, transform=None, output=None, validate=None, api=None, requires=(), map_params=None, **kwargs):
         self.route = {}
         if transform is not None:
             self.route['transform'] = transform
@@ -53,6 +54,8 @@ class Router(object):
             self.route['api'] = api
         if requires:
             self.route['requires'] = (requires, ) if not isinstance(requires, (tuple, list)) else requires
+        if map_params:
+            self.route['map_params'] = map_params
 
     def output(self, formatter, **overrides):
         """Sets the output formatter that should be used to render this route"""
@@ -81,6 +84,10 @@ class Router(object):
         return self.where(requires=tuple(set(self.route.get('requires', ())).difference(requirements if
                                                             type(requirements) in (list, tuple) else (requirements, ))))
 
+    def map_params(self, **map_params):
+        """Map interface specific params to an internal name representation"""
+        return self.where(map_params=map_params)
+
     def where(self, **overrides):
         """Creates a new route, based on the current route, with the specified overrided values"""
         route_data = self.route.copy()
@@ -183,10 +190,12 @@ class HTTPRouter(InternalValidation):
     """The HTTPRouter provides the base concept of a router from an HTTPRequest to a Python function"""
     __slots__ = ()
 
-    def __init__(self, versions=None, parse_body=False, parameters=None, defaults={}, status=None,
+    def __init__(self, versions=any, parse_body=False, parameters=None, defaults={}, status=None,
                  response_headers=None, private=False, inputs=None, **kwargs):
         super().__init__(**kwargs)
-        self.route['versions'] = (versions, ) if isinstance(versions, (int, float, None.__class__)) else versions
+        if versions is not any:
+            self.route['versions'] = (versions, ) if isinstance(versions, (int, float, None.__class__)) else versions
+            self.route['versions'] = tuple(int(version) if version else version for version in self.route['versions'])
         if parse_body:
             self.route['parse_body'] = parse_body
         if parameters:
@@ -244,25 +253,41 @@ class HTTPRouter(InternalValidation):
                  no_store and 'no-store', must_revalidate and 'must-revalidate')
         return self.add_response_headers({'cache-control': ', '.join(filter(bool, parts))}, **overrides)
 
-    def allow_origins(self, *origins, methods=None, **overrides):
+    def allow_origins(self, *origins, methods=None, max_age=None, credentials=None, headers=None, **overrides):
         """Convience method for quickly allowing other resources to access this one"""
-        headers = {'Access-Control-Allow-Origin': ', '.join(origins) if origins else '*'}
+        response_headers = {}
+        if origins:
+            @hug.response_middleware()
+            def process_data(request, response, resource):
+                if 'ORIGIN' in request.headers:
+                    origin = request.headers['ORIGIN']
+                    if origin in origins:
+                        response.set_header('Access-Control-Allow-Origin', origin)
+        else:
+            response_headers['Access-Control-Allow-Origin'] = '*'
+
         if methods:
-            headers['Access-Control-Allow-Methods'] = ', '.join(methods)
-        return self.add_response_headers(headers, **overrides)
+            response_headers['Access-Control-Allow-Methods'] = ', '.join(methods)
+        if max_age:
+            response_headers['Access-Control-Max-Age'] = max_age
+        if credentials:
+            response_headers['Access-Control-Allow-Credentials'] = str(credentials).lower()
+        if headers:
+            response_headers['Access-Control-Allow-Headers'] = headers
+        return self.add_response_headers(response_headers, **overrides)
 
 
 class NotFoundRouter(HTTPRouter):
     """Provides a chainable router that can be used to route 404'd request to a Python function"""
     __slots__ = ()
 
-    def __init__(self, output=None, versions=None, status=falcon.HTTP_NOT_FOUND, **kwargs):
+    def __init__(self, output=None, versions=any, status=falcon.HTTP_NOT_FOUND, **kwargs):
         super().__init__(output=output, versions=versions, status=status, **kwargs)
 
     def __call__(self, api_function):
         api = self.route.get('api', hug.api.from_object(api_function))
         (interface, callable_method) = self._create_interface(api, api_function)
-        for version in self.route['versions']:
+        for version in self.route.get('versions', (None, )):
             api.http.set_not_found_handler(interface, version)
 
         return callable_method
@@ -336,7 +361,7 @@ class ExceptionRouter(HTTPRouter):
     def __call__(self, api_function):
         api = self.route.get('api', hug.api.from_object(api_function))
         (interface, callable_method) = self._create_interface(api, api_function, catch_exceptions=False)
-        for version in self.route['versions']:
+        for version in self.route.get('versions', (None, )):
             for exception in self.route['exceptions']:
                 api.http.add_exception_handler(exception, interface, version)
 
@@ -351,7 +376,7 @@ class URLRouter(HTTPRouter):
     """Provides a chainable router that can be used to route a URL to a Python function"""
     __slots__ = ()
 
-    def __init__(self, urls=None, accept=HTTP_METHODS, output=None, examples=(), versions=None,
+    def __init__(self, urls=None, accept=HTTP_METHODS, output=None, examples=(), versions=any,
                  suffixes=(), prefixes=(), response_headers=None, parse_body=True, **kwargs):
         super().__init__(output=output, versions=versions, parse_body=parse_body, response_headers=response_headers,
                          **kwargs)
@@ -388,7 +413,7 @@ class URLRouter(HTTPRouter):
                 handlers = api.http.routes[api.http.base_url].setdefault(url, {})
                 for method in self.route.get('accept', ()):
                     version_mapping = handlers.setdefault(method.upper(), {})
-                    for version in self.route['versions']:
+                    for version in self.route.get('versions', (None, )):
                         version_mapping[version] = interface
                         api.http.versioned.setdefault(version, {})[callable_method.__name__] = callable_method
 
