@@ -23,11 +23,11 @@ from __future__ import absolute_import
 
 import uuid as native_uuid
 from decimal import Decimal
-from json import loads as load_json
 
 import hug._empty as empty
 from hug import introspect
 from hug.exceptions import InvalidTypeData
+from hug.json_module import json as json_converter
 
 
 class Type(object):
@@ -35,7 +35,8 @@ class Type(object):
        Override `__call__` to define how the type should be transformed and validated
     """
     _hug_type = True
-    __slots__ = ()
+    _sub_type = None
+    _accept_context = False
 
     def __init__(self):
         pass
@@ -44,52 +45,115 @@ class Type(object):
         raise NotImplementedError('To implement a new type __call__ must be defined')
 
 
-def create(doc=None, error_text=None, exception_handlers=empty.dict, extend=Type, chain=True, auto_instance=True):
+def create(doc=None, error_text=None, exception_handlers=empty.dict, extend=Type, chain=True, auto_instance=True,
+           accept_context=False):
     """Creates a new type handler with the specified type-casting handler"""
     extend = extend if type(extend) == type else type(extend)
 
     def new_type_handler(function):
         class NewType(extend):
             __slots__ = ()
+            _accept_context = accept_context
 
             if chain and extend != Type:
                 if error_text or exception_handlers:
-                    def __call__(self, value):
-                        try:
+                    if not accept_context:
+                        def __call__(self, value):
+                            try:
+                                value = super(NewType, self).__call__(value)
+                                return function(value)
+                            except Exception as exception:
+                                for take_exception, rewrite in exception_handlers.items():
+                                    if isinstance(exception, take_exception):
+                                        if isinstance(rewrite, str):
+                                            raise ValueError(rewrite)
+                                        else:
+                                            raise rewrite(value)
+                                if error_text:
+                                    raise ValueError(error_text)
+                                raise exception
+                    else:
+                        if extend._accept_context:
+                            def __call__(self, value, context):
+                                try:
+                                    value = super(NewType, self).__call__(value, context)
+                                    return function(value, context)
+                                except Exception as exception:
+                                    for take_exception, rewrite in exception_handlers.items():
+                                        if isinstance(exception, take_exception):
+                                            if isinstance(rewrite, str):
+                                                raise ValueError(rewrite)
+                                            else:
+                                                raise rewrite(value)
+                                    if error_text:
+                                        raise ValueError(error_text)
+                                    raise exception
+                        else:
+                            def __call__(self, value, context):
+                                try:
+                                    value = super(NewType, self).__call__(value)
+                                    return function(value, context)
+                                except Exception as exception:
+                                    for take_exception, rewrite in exception_handlers.items():
+                                        if isinstance(exception, take_exception):
+                                            if isinstance(rewrite, str):
+                                                raise ValueError(rewrite)
+                                            else:
+                                                raise rewrite(value)
+                                    if error_text:
+                                        raise ValueError(error_text)
+                                    raise exception
+                else:
+                    if not accept_context:
+                        def __call__(self, value):
                             value = super(NewType, self).__call__(value)
                             return function(value)
-                        except Exception as exception:
-                            for take_exception, rewrite in exception_handlers.items():
-                                if isinstance(exception, take_exception):
-                                    if isinstance(rewrite, str):
-                                        raise ValueError(rewrite)
-                                    else:
-                                        raise rewrite(value)
-                            if error_text:
-                                raise ValueError(error_text)
-                            raise exception
-                else:
-                    def __call__(self, value):
-                        value = super(NewType, self).__call__(value)
-                        return function(value)
+                    else:
+                        if extend._accept_context:
+                            def __call__(self, value, context):
+                                value = super(NewType, self).__call__(value, context)
+                                return function(value, context)
+                        else:
+                            def __call__(self, value, context):
+                                value = super(NewType, self).__call__(value)
+                                return function(value, context)
             else:
-                if error_text or exception_handlers:
-                    def __call__(self, value):
-                        try:
+                if not accept_context:
+                    if error_text or exception_handlers:
+                        def __call__(self, value):
+                            try:
+                                return function(value)
+                            except Exception as exception:
+                                for take_exception, rewrite in exception_handlers.items():
+                                    if isinstance(exception, take_exception):
+                                        if isinstance(rewrite, str):
+                                            raise ValueError(rewrite)
+                                        else:
+                                            raise rewrite(value)
+                                if error_text:
+                                    raise ValueError(error_text)
+                                raise exception
+                    else:
+                        def __call__(self, value):
                             return function(value)
-                        except Exception as exception:
-                            for take_exception, rewrite in exception_handlers.items():
-                                if isinstance(exception, take_exception):
-                                    if isinstance(rewrite, str):
-                                        raise ValueError(rewrite)
-                                    else:
-                                        raise rewrite(value)
-                            if error_text:
-                                raise ValueError(error_text)
-                            raise exception
                 else:
-                    def __call__(self, value):
-                        return function(value)
+                    if error_text or exception_handlers:
+                        def __call__(self, value, context):
+                            try:
+                                return function(value, context)
+                            except Exception as exception:
+                                for take_exception, rewrite in exception_handlers.items():
+                                    if isinstance(exception, take_exception):
+                                        if isinstance(rewrite, str):
+                                            raise ValueError(rewrite)
+                                        else:
+                                            raise rewrite(value)
+                                if error_text:
+                                    raise ValueError(error_text)
+                                raise exception
+                    else:
+                        def __call__(self, value, context):
+                            return function(value, context)
 
         NewType.__doc__ = function.__doc__ if doc is None else doc
         if auto_instance and not (introspect.arguments(NewType.__init__, -1) or
@@ -101,9 +165,15 @@ def create(doc=None, error_text=None, exception_handlers=empty.dict, extend=Type
     return new_type_handler
 
 
-def accept(kind, doc=None, error_text=None, exception_handlers=empty.dict):
+def accept(kind, doc=None, error_text=None, exception_handlers=empty.dict, accept_context=False):
     """Allows quick wrapping of any Python type cast function for use as a hug type annotation"""
-    return create(doc, error_text, exception_handlers=exception_handlers, chain=False)(kind)
+    return create(
+        doc,
+        error_text,
+        exception_handlers=exception_handlers,
+        chain=False,
+        accept_context=accept_context
+    )(kind)
 
 number = accept(int, 'A Whole number', 'Invalid whole number provided')
 float_number = accept(float, 'A float number', 'Invalid float number provided')
@@ -124,19 +194,29 @@ class Text(Type):
 text = Text()
 
 
-class Multiple(Type):
+class SubTyped(type):
+    def __getitem__(cls, sub_type):
+        class TypedSubclass(cls):
+            _sub_type = sub_type
+        return TypedSubclass
+
+
+class Multiple(Type, metaclass=SubTyped):
     """Multiple Values"""
     __slots__ = ()
 
     def __call__(self, value):
-        return value if isinstance(value, list) else [value]
+        as_multiple = value if isinstance(value, list) else [value]
+        if self._sub_type:
+            return [self._sub_type(item) for item in as_multiple]
+        return as_multiple
 
 
-class DelimitedList(Type):
+
+class DelimitedList(Type, metaclass=SubTyped):
     """Defines a list type that is formed by delimiting a list with a certain character or set of characters"""
-    __slots__ = ('using', )
-
     def __init__(self, using=","):
+        super().__init__()
         self.using = using
 
     @property
@@ -144,7 +224,10 @@ class DelimitedList(Type):
         return '''Multiple values, separated by "{0}"'''.format(self.using)
 
     def __call__(self, value):
-        return value if type(value) in (list, tuple) else value.split(self.using)
+        value_list = value if type(value) in (list, tuple) else value.split(self.using)
+        if self._sub_type:
+            value_list = [self._sub_type(val) for val in value_list]
+        return value_list
 
 
 class SmartBoolean(type(boolean)):
@@ -164,12 +247,26 @@ class SmartBoolean(type(boolean)):
         raise KeyError('Invalid value passed in for true/false field')
 
 
-class InlineDictionary(Type):
+class InlineDictionary(Type, metaclass=SubTyped):
     """A single line dictionary, where items are separted by commas and key:value are separated by a pipe"""
-    __slots__ = ()
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.key_type = self.value_type = None
+        if self._sub_type:
+            if type(self._sub_type) in (tuple, list):
+                if len(self._sub_type) >= 2:
+                    self.key_type, self.value_type = self._sub_type[:2]
+            else:
+                self.key_type = self._sub_type
 
     def __call__(self, string):
-        return {key.strip(): value.strip() for key, value in (item.split(":") for item in string.split("|"))}
+        dictionary = {}
+        for key, value in (item.split(":") for item in string.split("|")):
+            key, value = key.strip(), value.strip()
+            dictionary[self.key_type(key)
+                       if self.key_type else key] = self.value_type(value) if self.value_type else value
+        return dictionary
 
 
 class OneOf(Type):
@@ -214,7 +311,14 @@ class JSON(Type):
     def __call__(self, value):
         if type(value) in (str, bytes):
             try:
-                return load_json(value)
+                return json_converter.loads(value)
+            except Exception:
+                raise ValueError('Incorrectly formatted JSON provided')
+        if type(value) is list:
+            # If Falcon is set to comma-separate entries, this segment joins them again.
+            try:
+                fixed_value = ",".join(value)
+                return json_converter.loads(fixed_value)
             except Exception:
                 raise ValueError('Incorrectly formatted JSON provided')
         else:
@@ -304,7 +408,7 @@ class GreaterThan(Type):
 
 
 class Length(Type):
-    """Accepts a a value that is withing a specific length limit"""
+    """Accepts a a value that is within a specific length limit"""
     __slots__ = ('lower', 'upper', 'convert')
 
     def __init__(self, lower, upper, convert=text):
@@ -447,7 +551,6 @@ class NewTypeMeta(type):
 
 class Schema(object, metaclass=NewTypeMeta):
     """Schema for creating complex types using hug types"""
-    _hug_type = True
     __slots__ = ()
 
     def __new__(cls, json, *args, **kwargs):
@@ -466,9 +569,9 @@ class Schema(object, metaclass=NewTypeMeta):
 json = JSON()
 
 
-class MarshmallowSchema(Type):
-    """Allows using a Marshmallow Schema directly in a hug type annotation"""
-    __slots__ = ("schema", )
+class MarshmallowInputSchema(Type):
+    """Allows using a Marshmallow Schema directly in a hug input type annotation"""
+    __slots__ = ("schema")
 
     def __init__(self, schema):
         self.schema = schema
@@ -477,13 +580,41 @@ class MarshmallowSchema(Type):
     def __doc__(self):
         return self.schema.__doc__ or self.schema.__class__.__name__
 
-    def __call__(self, value):
+    def __call__(self, value, context):
+        self.schema.context = context
         value, errors = self.schema.loads(value) if isinstance(value, str) else self.schema.load(value)
         if errors:
             raise InvalidTypeData('Invalid {0} passed in'.format(self.schema.__class__.__name__), errors)
         return value
 
 
+class MarshmallowReturnSchema(Type):
+    """Allows using a Marshmallow Schema directly in a hug return type annotation"""
+    __slots__ = ("schema", )
+
+    def __init__(self, schema):
+        self.schema = schema
+
+    @property
+    def context(self):
+        return self.schema.context
+
+    @context.setter
+    def context(self, context):
+        self.schema.context = context
+
+    @property
+    def __doc__(self):
+        return self.schema.__doc__ or self.schema.__class__.__name__
+
+    def __call__(self, value):
+        value, errors = self.schema.dump(value)
+        if errors:
+            raise InvalidTypeData('Invalid {0} passed in'.format(self.schema.__class__.__name__), errors)
+        return value
+
+
+
 multiple = Multiple()
 smart_boolean = SmartBoolean()
 inline_dictionary = InlineDictionary()
