@@ -23,13 +23,17 @@ import json
 import os
 import sys
 from unittest import mock
+from collections import namedtuple
 
 import falcon
-import hug
 import pytest
 import requests
 from falcon.testing import StartResponseMock, create_environ
+from marshmallow import Schema, fields
+
+import hug
 from hug._async import coroutine
+from hug.exceptions import InvalidTypeData
 
 from .constants import BASE_DIRECTORY
 
@@ -54,18 +58,18 @@ def test_basic_call():
     assert hug.test.get(module, '/hello_world').data == "Hello World!"
 
 
-def test_basic_call_on_method():
+def test_basic_call_on_method(hug_api):
     """Test to ensure the most basic call still works if applied to a method"""
     class API(object):
 
-        @hug.call()
+        @hug.call(api=hug_api)
         def hello_world(self=None):
             return "Hello World!"
 
     api_instance = API()
     assert api_instance.hello_world.interface.http
     assert api_instance.hello_world() == 'Hello World!'
-    assert hug.test.get(api, '/hello_world').data == "Hello World!"
+    assert hug.test.get(hug_api, '/hello_world').data == "Hello World!"
 
     class API(object):
 
@@ -74,17 +78,17 @@ def test_basic_call_on_method():
 
     api_instance = API()
 
-    @hug.call()
+    @hug.call(api=hug_api)
     def hello_world():
         return api_instance.hello_world()
 
     assert api_instance.hello_world() == 'Hello World!'
-    assert hug.test.get(api, '/hello_world').data == "Hello World!"
+    assert hug.test.get(hug_api, '/hello_world').data == "Hello World!"
 
     class API(object):
 
         def __init__(self):
-            hug.call()(self.hello_world_method)
+            hug.call(api=hug_api)(self.hello_world_method)
 
         def hello_world_method(self):
             return "Hello World!"
@@ -92,12 +96,12 @@ def test_basic_call_on_method():
     api_instance = API()
 
     assert api_instance.hello_world_method() == 'Hello World!'
-    assert hug.test.get(api, '/hello_world_method').data == "Hello World!"
+    assert hug.test.get(hug_api, '/hello_world_method').data == "Hello World!"
 
 
-def test_single_parameter():
+def test_single_parameter(hug_api):
     """Test that an api with a single parameter interacts as desired"""
-    @hug.call()
+    @hug.call(api=hug_api)
     def echo(text):
         return text
 
@@ -106,8 +110,8 @@ def test_single_parameter():
     with pytest.raises(TypeError):
         echo()
 
-    assert hug.test.get(api, 'echo', text="Hello").data == "Hello"
-    assert 'required' in hug.test.get(api, '/echo').data['errors']['text'].lower()
+    assert hug.test.get(hug_api, 'echo', text="Hello").data == "Hello"
+    assert 'required' in hug.test.get(hug_api, '/echo').data['errors']['text'].lower()
 
 
 def test_on_invalid_transformer():
@@ -356,10 +360,19 @@ def test_versioning():
     def echo(text, api_version):
         return api_version
 
+    @hug.get('/echo', versions='8')  # noqa
+    def echo(text, api_version):
+        return api_version
+
     @hug.get('/echo', versions=False)  # noqa
     def echo(text):
         return "No Versions"
 
+    with pytest.raises(ValueError):
+        @hug.get('/echo', versions='eight')  # noqa
+        def echo(text, api_version):
+            return api_version
+
     assert hug.test.get(api, 'v1/echo', text="hi").data == 'hi'
     assert hug.test.get(api, 'v2/echo', text="hi").data == "Echo: hi"
     assert hug.test.get(api, 'v3/echo', text="hi").data == "Echo: hi"
@@ -367,6 +380,7 @@ def test_versioning():
     assert hug.test.get(api, 'echo', text="hi", headers={'X-API-VERSION': '3'}).data == "Echo: hi"
     assert hug.test.get(api, 'v4/echo', text="hi").data == "Not Implemented"
     assert hug.test.get(api, 'v7/echo', text="hi").data == 7
+    assert hug.test.get(api, 'v8/echo', text="hi").data == 8
     assert hug.test.get(api, 'echo', text="hi").data == "No Versions"
     assert hug.test.get(api, 'echo', text="hi", api_version=3, body={'api_vertion': 4}).data == "Echo: hi"
 
@@ -517,9 +531,13 @@ def test_custom_deserializer_support():
 
 def test_marshmallow_support():
     """Ensure that you can use Marshmallow style objects to control input and output validation and transformation"""
+    MarshalResult = namedtuple('MarshalResult', ['data', 'errors'])
+
     class MarshmallowStyleObject(object):
         def dump(self, item):
-            return 'Dump Success'
+            if item == 'bad':
+                return MarshalResult('', 'problems')
+            return MarshalResult('Dump Success', {})
 
         def load(self, item):
             return ('Load Success', None)
@@ -537,6 +555,14 @@ def test_marshmallow_support():
     assert test_marshmallow_style() == 'world'
 
 
+    @hug.get()
+    def test_marshmallow_style_error() -> schema:
+        return 'bad'
+
+    with pytest.raises(InvalidTypeData):
+        hug.test.get(api, 'test_marshmallow_style_error')
+
+
     @hug.get()
     def test_marshmallow_input(item: schema):
         return item
@@ -728,7 +754,13 @@ def test_extending_api():
         import tests.module_fake
         return (tests.module_fake, )
 
+    @hug.get('/fake/error')
+    def my_error():
+        import tests.module_fake
+        raise tests.module_fake.FakeException()
+
     assert hug.test.get(api, 'fake/made_up_api').data
+    assert hug.test.get(api, 'fake/error').data == True
 
 
 def test_extending_api_simple():
@@ -917,6 +949,69 @@ def test_cli_with_directives():
     assert isinstance(hug.test.cli(test), float)
 
 
+def test_cli_with_class_directives():
+
+    @hug.directive()
+    class ClassDirective(object):
+
+        def __init__(self, *args, **kwargs):
+            self.test = 1
+
+    @hug.cli()
+    @hug.local(skip_directives=False)
+    def test(class_directive: ClassDirective):
+        return class_directive.test
+
+    assert test() == 1
+    assert hug.test.cli(test) == 1
+
+    class TestObject(object):
+        is_cleanup_launched = False
+        last_exception = None
+
+    @hug.directive()
+    class ClassDirectiveWithCleanUp(object):
+
+        def __init__(self, *args, **kwargs):
+            self.test_object = TestObject
+
+        def cleanup(self, exception):
+            self.test_object.is_cleanup_launched = True
+            self.test_object.last_exception = exception
+
+    @hug.cli()
+    @hug.local(skip_directives=False)
+    def test2(class_directive: ClassDirectiveWithCleanUp):
+        return class_directive.test_object.is_cleanup_launched
+
+    assert not hug.test.cli(test2)  # cleanup should be launched after running command
+    assert TestObject.is_cleanup_launched
+    assert TestObject.last_exception is None
+    TestObject.is_cleanup_launched = False
+    TestObject.last_exception = None
+    assert not test2()
+    assert TestObject.is_cleanup_launched
+    assert TestObject.last_exception is None
+
+    @hug.cli()
+    @hug.local(skip_directives=False)
+    def test_with_attribute_error(class_directive: ClassDirectiveWithCleanUp):
+        raise class_directive.test_object2
+
+    hug.test.cli(test_with_attribute_error)
+    assert TestObject.is_cleanup_launched
+    assert isinstance(TestObject.last_exception, AttributeError)
+    TestObject.is_cleanup_launched = False
+    TestObject.last_exception = None
+    try:
+        test_with_attribute_error()
+        assert False
+    except AttributeError:
+        assert True
+    assert TestObject.is_cleanup_launched
+    assert isinstance(TestObject.last_exception, AttributeError)
+
+
 def test_cli_with_named_directives():
     """Test to ensure you can pass named directives into the cli"""
     @hug.cli()
@@ -1200,6 +1295,12 @@ def test_cli_with_multiple_ints():
 
     assert hug.test.cli(test_multiple_cli, ints=['1', '2', '3']) == [1, 2, 3]
 
+    @hug.cli()
+    def test_multiple_cli(ints: hug.types.Multiple[int]()=[]):
+        return ints
+
+    assert hug.test.cli(test_multiple_cli, ints=['1', '2', '3']) == [1, 2, 3]
+
 
 def test_startup():
     """Test to ensure hug startup decorators work as expected"""
@@ -1454,9 +1555,43 @@ def test_cli_kwargs(hug_api):
 
 
 def test_api_gets_extra_variables_without_kargs_or_kwargs(hug_api):
+    """Test to ensure it's possiible to extra all params without specifying them exactly"""
     @hug.get(api=hug_api)
     def ensure_params(request, response):
         return request.params
 
     assert hug.test.get(hug_api, 'ensure_params', params={'make': 'it'}).data == {'make': 'it'}
     assert hug.test.get(hug_api, 'ensure_params', hello='world').data == {'hello': 'world'}
+
+
+def test_utf8_output(hug_api):
+    """Test to ensure unicode data is correct outputed on JSON outputs without modification"""
+    @hug.get(api=hug_api)
+    def output_unicode():
+        return {'data': 'Τη γλώσσα μου έδωσαν ελληνική'}
+
+    assert hug.test.get(hug_api, 'output_unicode').data == {'data': 'Τη γλώσσα μου έδωσαν ελληνική'}
+
+
+def test_param_rerouting(hug_api):
+    @hug.local(api=hug_api, map_params={'local_id': 'record_id'})
+    @hug.cli(api=hug_api, map_params={'cli_id': 'record_id'})
+    @hug.get(api=hug_api, map_params={'id': 'record_id'})
+    def pull_record(record_id: hug.types.number):
+        return record_id
+
+    assert hug.test.get(hug_api, 'pull_record', id=10).data == 10
+    assert hug.test.get(hug_api, 'pull_record', id='10').data == 10
+    assert 'errors' in hug.test.get(hug_api, 'pull_record', id='ten').data
+    assert hug.test.cli(pull_record, cli_id=10) == 10
+    assert hug.test.cli(pull_record, cli_id='10') == 10
+    with pytest.raises(SystemExit):
+        hug.test.cli(pull_record, cli_id='ten')
+    assert pull_record(local_id=10)
+
+    @hug.get(api=hug_api, map_params={'id': 'record_id'})
+    def pull_record(record_id: hug.types.number=1):
+        return record_id
+
+    assert hug.test.get(hug_api, 'pull_record').data == 1
+    assert hug.test.get(hug_api, 'pull_record', id=10).data == 10
